{% set info = h.files_link_details(data[field.field_name]) %}
{% if info %}
    {% if info.href %}
        <a href="{{ info.href }}" data-files-content-type="{{ info.content_type }}">{{ info.label }}</a>
    {% else %}
        {{ info.label }}
    {% endif %}

{% endif %}
